# authentic2 - versatile identity manager
# Copyright (C) 2010-2018 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import transaction

from rest_framework.serializers import raise_errors_on_nested_writes
from rest_framework.utils import model_meta


class GetOrCreateModelSerializer(object):
    def get_or_create(self, keys, validated_data):
        raise_errors_on_nested_writes('get_or_create', self, validated_data)

        ModelClass = self.Meta.model

        # Remove many-to-many relationships from validated_data.
        # They are not valid arguments to the default `.create()` method,
        # as they require that the instance has already been saved.
        info = model_meta.get_field_info(ModelClass)
        many_to_many = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        kwargs = {}
        defaults = kwargs['defaults'] = {}
        missing_keys = set(keys) - set(validated_data)
        if missing_keys:
            raise TypeError('Keys %s are missing' % missing_keys)
        for key, value in validated_data.items():
            if key in keys:
                kwargs[key] = value
            else:
                defaults[key] = value
        with transaction.atomic():
            instance, created = self.Meta.model._default_manager.get_or_create(**kwargs)
            if many_to_many and created:
                self.update(instance, many_to_many)
        return instance

    def update_or_create(self, keys, validated_data):
        raise_errors_on_nested_writes('update_or_create', self, validated_data)

        ModelClass = self.Meta.model

        # Remove many-to-many relationships from validated_data.
        # They are not valid arguments to the default `.create()` method,
        # as they require that the instance has already been saved.
        info = model_meta.get_field_info(ModelClass)
        many_to_many = {}
        get_or_create_data = validated_data.copy()
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = get_or_create_data.pop(field_name)

        kwargs = {}
        defaults = kwargs['defaults'] = {}
        missing_keys = set(keys) - set(get_or_create_data)
        if missing_keys:
            raise TypeError('Keys %s are missing' % missing_keys)
        for key, value in get_or_create_data.items():
            if key in keys:
                kwargs[key] = value
            else:
                defaults[key] = value
        with transaction.atomic():
            instance, created = self.Meta.model._default_manager.get_or_create(**kwargs)
            if many_to_many or not created:
                self.update(instance, validated_data)
        return instance

    def create(self, validated_data):
        try:
            keys = self.context['view'].request.GET.getlist('get_or_create')
        except Exception:
            pass
        else:
            if keys:
                return self.get_or_create(keys, validated_data)
        try:
            keys = self.context['view'].request.GET.getlist('update_or_create')
        except Exception:
            pass
        else:
            if keys:
                return self.update_or_create(keys, validated_data)
        return super(GetOrCreateModelSerializer, self).create(validated_data)
